using System.Diagnostics;
using System.Text;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.CSharp;
using Microsoft.CodeAnalysis.CSharp.Syntax;
using Microsoft.CodeAnalysis.Text;
using TypeInfo = Microsoft.CodeAnalysis.TypeInfo;

namespace ServerDatabase.SourceGenerator;

[Generator]
public class TableRowGenerator : ISourceGenerator
{
    public void Initialize(GeneratorInitializationContext context)
    {
#if DEBUG
        if (!Debugger.IsAttached) Debugger.Launch();
#endif
    }
    
    public void Execute(GeneratorExecutionContext context)
    {
        try
        {
            INamedTypeSymbol? attributeSymbol = context.Compilation.GetTypeByMetadataName(typeof(TableRowAttribute).FullName);

            IEnumerable<SyntaxTree> classWithAttributes = context.Compilation.SyntaxTrees.Where(st => st.GetRoot().DescendantNodes().OfType<ClassDeclarationSyntax>()
                .Any(p => p.DescendantNodes().OfType<AttributeSyntax>().Any()));
            
            foreach (SyntaxTree? tree in classWithAttributes)
            {
                if (tree is null) continue;
                SemanticModel semanticModel = context.Compilation.GetSemanticModel(tree);
                foreach (ClassDeclarationSyntax? declaredClass in tree.GetRoot().DescendantNodes().OfType<ClassDeclarationSyntax>().Where(cd => cd.DescendantNodes().OfType<AttributeSyntax>().Any()))
                {
                    if (declaredClass is null) continue;
                    List<SyntaxToken>? nodes = declaredClass
                        .DescendantNodes()
                        .OfType<AttributeSyntax>()
                        .FirstOrDefault(a => a.DescendantTokens().Any(dt => dt.IsKind(SyntaxKind.IdentifierToken) && dt.Parent is not null && semanticModel.GetTypeInfo(dt.Parent).Type is not null && semanticModel.GetTypeInfo(dt.Parent).Type.Name == attributeSymbol.Name))
                        ?.DescendantTokens()
                        ?.Where(dt => dt.IsKind(SyntaxKind.IdentifierToken))
                        ?.ToList();
                    
                    if(nodes == null)
                    {
                        continue;
                    }
                    
                    TypeInfo relatedClass = semanticModel.GetTypeInfo(nodes.Last().Parent);

                    StringBuilder generatedClass = this.GenerateClass(tree.GetRoot().DescendantNodes().OfType<FileScopedNamespaceDeclarationSyntax>().First().Name.ToString(), declaredClass);

                    foreach(IPropertySymbol? classProperty in relatedClass.Type.GetMembers().OfType<IPropertySymbol>())
                    {
                        this.GenerateProperty(classProperty, ref generatedClass);
                    }

                    this.CloseClass(generatedClass);

                    context.AddSource($"{declaredClass.Identifier}_{relatedClass.Type.Name}.g.cs", SourceText.From(generatedClass.ToString(), Encoding.UTF8));
                    //context.AddSource($"{declaredClass.Identifier}_{relatedClass.Type.Name}_goooddddd.g.cs", SourceText.From(generatedClass.ToString().Replace("partial ", "").Replace(declaredClass.Identifier.ToString(), declaredClass.Identifier.ToString() + "GODDD"), Encoding.UTF8));
                }
            }
        }
        catch (Exception e)
        {
            context.AddSource("teste.g.cs", SourceText.From(@$"// <auto-generated/>
using System;
namespace ServerDatabase.SourceGenerator.Generated
{{
    public class bobe
    {{
        public string e = @""{e}"";
    }}
}}", Encoding.UTF8));
        }
        return;
    }
    
    private void GenerateProperty(IPropertySymbol prop, ref StringBuilder builder)
    {
        string bad = prop.Type.ToString();
        bad = bad.Replace("ServerDatabase.TableColumn<", "");
        bad = bad.Remove(bad.LastIndexOf('>'));
        builder.AppendLine($"       public {bad} {prop.Name} {{ get; set; }}");
    }
    
    private StringBuilder GenerateClass(string n, ClassDeclarationSyntax og)
    {
        StringBuilder sb = new StringBuilder();
        sb.Append(@$"// <auto-generated/>
using System;

namespace {n}
{{
    public partial class " + og.Identifier);

        sb.Append(@"
    {
");

        return sb;
    }
    
    private void CloseClass(StringBuilder generatedClass)
    {
        generatedClass.Append(
            @"    }
}");
    }
}